from .property import Property
from .properties import Properties
from .endpoints import Endpoints
from .component import Component
from .instance import Instance
from .component_test import ComponentTest



import random
import uuid
import exrex
import json

class Message:
    def __init__(self, msg):
        self.msg = msg

def rand_str(min_length=5, max_length=20):
    """Return a random non-empty string."""
    if min_length <= 0:
        raise ValueError("min_length must be greater than 0.")
    if min_length > max_length:
        raise ValueError("min_length must by greater than or equal to max_length.")
    length = random.randint(min_length-1, max_length-1)
    cfg = "[a-z A-Z 0-9][a-z A-Z 0-9 ]{{{}}}".format(length)
    return exrex.getone(cfg)

def rand_int(min_value=1, max_value=1000):
    """Return a randon int."""
    if min_value > max_value:
        raise ValueError("max_length must be greater than or equal to min_length.")
    return random.randint(min_value, max_value)

def rand_uuid(as_string=True):
    """Return a random uuid."""
    return str(uuid.uuid4()) if as_string else uuid.uuid4()


def value_from_dict(dict_, key):
    """Return the value for the key.

    Parameters
    ----------
    dict_: dict
        Dict containing the key
    key: str
        Key to lookup in the dict
    """
    return dict_[key]

def dict_has_value(dict_, key):
    """Return True if the key is contained in the dict.

    Parameters
    ----------
    dict_: dict
        Dict to check in
    key: str
        Key to check
    """
    return key in dict_

def attr_from_obj(obj, name):
    """Return the value of the attribute name.

    Parameters
    ----------
    obj: object
        Object with attribute name
    name: str
        Attribute name to return
    """
    return getattr(obj, name)

def obj_has_attr(obj, name):
    """Return True if the object has the attribute.

    Parameters
    ----------
    obj: object
        Object to check
    name: str
        Attribute name to check
    """
    return hasattr(obj, name)



class Component:
    """Specification of a resource managed by an API.

    Parameters
    ----------
    name: str
        Name of the component
    props: Properties object
        Properties of the component
    endpoints: Endpoints object
        Endpoints of this component
    parent: Component object, default None
        Parent component

    Examples
    --------
    prop1 = Property("prop1", rand_str, rand_int)
    prop2 = Property("prop2", rand_int, rand_str)
    props = Properties(prop1, prop2)
    component_api = Endpoints(
        create=lambda values,parent: {**values},
        delete=lambda values,instance: {**instance.data}
    )
    comp = Component("Test", props, component_api)
    """
    def __init__(self, name, props, endpoints, parent=None):
        self.name = name
        self.props = props
        self.endpoints = endpoints
        self.parent = parent

    def get_valids(self, endpoint, count=1):
        """Generate and return list of valids from properties for endpoint."""
        return self.props.get_valids(endpoint, count)

    def get_invalids(self, endpoint, count=1):
        """Generate and return list of invalids from properties for endpoint."""
        return self.props.get_invalids(endpoint, count)


import pytest
from .util import rand_int, rand_str, rand_uuid
from .property import Property
from .properties import Properties
from .endpoints import Endpoints

def test_constr():
    p1 = Property("p1", rand_str, rand_int)
    p2 = Property("p2", rand_int, rand_str)
    p3 = Property("p3", rand_uuid, rand_str)
    c1 = Component("c1", Properties(p1, p2), Endpoints())
    c2 = Component("c2", Properties(p1), Endpoints(), c1)



from .endpoints import Endpoints

class Property:
    """Model of a single property with different valid and invalid values.

    Parameters
    ----------
    name: str
        Name of the property
    valids: list or callable
        Specification of valid values
    invalids: list or callable
        Specification of invalid values
    required: list of str, default None
        List of endpoints this property is required for
    excluded: list of str, default ['GET', DELETE', 'LIST']
        List of endpoints this property should be excluded from
    spec_name: str, default None
        Name used in the payload spec (name if None)
    res_name: str, default None
        Name expected in the endpoint's result (nane if None)

    Examples
    --------
    Property("name", rand_str, [""], [Endpoints.CREATE])
    Property("description", rand_str, [])
    Property("count", rand_int, [])
    """
    def __init__(self, name, valids, invalids,
                 required=None, excluded=None,
                 spec_name=None, res_name=None):
        self.name = name
        self.valids = valids
        self.invalids = invalids
        self.required = required if required is not None else []
        self.excluded = excluded if excluded is not None else [Endpoints.GET, Endpoints.DELETE, Endpoints.LIST]
        for r in self.required:
            if r in self.excluded:
                raise ValueError("'{}' cannot be required AND excluded endpoint".format(r))
        self.spec_name = spec_name if spec_name is not None else name
        self.res_name = res_name if res_name is not None else name

    def __str__(self):
        return self.name

    def is_required(self, endpoint):
        """Return True if this property is required for the endpoint."""
        return endpoint in self.required

    def is_excluded(self, endpoint):
        """Return True if this property should be excluded for the endpoint."""
        return endpoint in self.excluded

    def _get_values(self, endpoint, values, count):
        """Return a list of vlud for the endpoint.

        Returns an empty list if this property is excluded for
        the endpoint.
        If values is specified as a list, a copy of this list
        is returned (i.e. count is ignored).
        If values is specified as a callable, count values
        are generated and returned in a new list.

        Parameters
        ----------
        endpoint: str
            Key of the endpoint to get values for
        values: list or callable
            Specification of the values
        count: int
            Number of values to generate if values is callable
        """
        if self.is_excluded(endpoint):
            return []
        if callable(values):
            values = [values() for _ in range(count)]
        else:
            values = [v for v in values]
        return values

    def get_valids(self, endpoint, count=1):
        """Return a list of count valid values of this property for the endpoint.

        If valids is specified as a list, a copy of this list
        is returned (i.e. count is ignored).
        If valids is specified as a callable, count values
        are generated and returned in a new list.
        If this property is not required for the endpoint,
        None is added to the list.

        endpoint: str
            Key of the endpoint to get values for
        count: int
            Number of values to generate if valid values are specified as callable
        """
        values = self._get_values(endpoint, self.valids, count)
        if not self.is_required(endpoint):
            values.append(None)
        return values

    def get_invalids(self, endpoint, count=1):
        """Return a list of count invalid values of this property for the endpoint.

        If invalids is specified as a list, a copy of this list
        is returned (i.e. count is ignored).
        If invalids is specified as a callable, count values
        are generated and returned in a new list.
        If this property is required for the endpoint,
        None is added to the list.

        endpoint: str
            Key of the endpoint to get values for
        count: int
            Number of values to generate if valid values are specified as callable
        """
        values = self._get_values(endpoint, self.invalids, count)
        if self.is_required(endpoint):
            values.append(None)
        return values

import pytest
from .util import rand_str, rand_int

def test_constr():
    with pytest.raises(Exception):
        Property()
    with pytest.raises(Exception):
        Property("")
    with pytest.raises(Exception):
        Property("", [])

def test_names():
    name, spec_name, res_name = rand_str(), rand_str(), rand_str()
    prop = Property(name, [], [])
    assert prop.name == name
    assert prop.spec_name == name
    assert prop.res_name == name
    prop = Property(name, [], [], spec_name=spec_name)
    assert prop.name == name
    assert prop.spec_name == spec_name
    assert prop.res_name == name
    prop = Property(name, [], [], res_name=res_name)
    assert prop.name == name
    assert prop.spec_name == name
    assert prop.res_name == res_name
    prop = Property(name, [], [], spec_name=spec_name, res_name=res_name)
    assert prop.name == name
    assert prop.spec_name == spec_name
    assert prop.res_name == res_name

def test_required():
    prop = Property("prop", [], [], required=["a", "b", "c"])
    assert prop.is_required("b")
    assert not prop.is_required("d")

def test_excluded():
    prop = Property("prop", [], [], excluded=["a", "b", "c"])
    assert prop.is_excluded("b")
    assert not prop.is_excluded("d")

def test_required_and_excluded():
    Property("prop", [], [], required=[], excluded=[])
    Property("prop", [], [], required=["a", "d"], excluded=["b", "c"])
    with pytest.raises(Exception):
        Property("prop", [], [], required=["a"], excluded=["a"])
    with pytest.raises(Exception):
        Property("prop", [], [], required=["a", "b"], excluded=["c", "d", "b"])

def test_valid():
    valids = [1, 3, 5]
    prop = Property("prop", valids, None, [Endpoints.CREATE])
    assert None not in prop.get_valids(Endpoints.CREATE)
    assert None in prop.get_valids(Endpoints.UPDATE)
    assert set(valids) == set(prop.get_valids(Endpoints.CREATE))

def test_invalid():
    invalids = ["a", "b"]
    prop = Property("prop", None, invalids, [Endpoints.CREATE])
    assert None in prop.get_invalids(Endpoints.CREATE)
    assert None not in prop.get_invalids(Endpoints.UPDATE)
    assert set(invalids) == set(prop.get_invalids(""))

def test_valid_counts():
    valids = [1, 3, 5]
    prop = Property("prop", valids, None, [Endpoints.CREATE])
    assert len(prop.get_valids(Endpoints.CREATE)) == 3
    assert len(prop.get_valids(Endpoints.CREATE, count=100)) == 3
    assert len(prop.get_valids(Endpoints.UPDATE)) == 4
    assert len(prop.get_valids(Endpoints.UPDATE, count=100)) == 4
    prop = Property("prop", rand_int, None, [Endpoints.CREATE])
    assert len(prop.get_valids(Endpoints.CREATE, count=100)) == 100
    assert len(prop.get_valids(Endpoints.UPDATE, count=100)) == 101

def test_invalid_counts():
    invalids = ["a", "b"]
    prop = Property("prop", None, invalids, [Endpoints.CREATE])
    assert len(prop.get_invalids(Endpoints.CREATE)) == 3
    assert len(prop.get_invalids(Endpoints.CREATE, count=100)) == 3
    assert len(prop.get_invalids(Endpoints.UPDATE)) == 2
    assert len(prop.get_invalids(Endpoints.UPDATE, count=100)) == 2
    prop = Property("prop", None, rand_str, [Endpoints.CREATE])
    assert len(prop.get_invalids(Endpoints.CREATE, count=100)) == 101
    assert len(prop.get_invalids(Endpoints.UPDATE, count=100)) == 100



from .endpoints import Endpoints
from .instance import Instance
import pytest

class ComponentTest:
    """Execution of common test patterns for a component.

    ComponentTest provides standardized tests for all
    CRUDL operations, specified as endpoints for the
    component.
    In addition, arbitrary endpoints can be tested
    with custom validation methods.

    Parameters
    ----------
    comp: Component object
        Component to execute tests for

    Examples
    --------
    comp = ...
    test = ComponentTest(comp)
    test.test_create(2)
    test.test_update_invalid()
    test.test_update_valid(3)
    """
    def __init__(self, comp):
        self.comp = comp

    def test_create_valid(self, count=1, exp_props=None):
        """Test the success and result of valid input to the create endpoint.

        This test creates instances of the component using
        valid input data, generated by the component.

        For each component instance, it checks if all properties,
        specified in valids, are also present in the endpoints
        result and their values match.
        You can limit the set of tested properties by specifying a list
        of expected properties (exp_props), specified by their name.
        """
        print("# # # test create (valid)")
        for valids in self.comp.get_valids(Endpoints.CREATE, count):
            exp_props_ = [prop.name for prop in self.comp.props.props if prop.spec_name in valids] if exp_props is None else exp_props
            print("valids:", valids)
            with Instance(self.comp, **valids) as instance:
                for name in exp_props_:
                    print("   ", name)
                    p = self.comp.props.get_prop(name)
                    assert self.comp.endpoints.res_has_prop(instance.data, p.res_name)
                    assert valids[p.spec_name] == self.comp.endpoints.prop_from_res(instance.data, p.res_name)

    def test_create_invalid(self, count=1):
        """Test the success of invalid input to the create endpoint.

        This test attempts to create instance of the component
        using invalid data, generated by the component.

        For each creation attempt, it checks if the invalid creation
        indeed caused raising an exception.
        """
        print("# # # test create (invalid)")
        for invalids in self.comp.get_invalids(Endpoints.CREATE, count):
            print("invalids:", invalids)
            with pytest.raises(Exception) as e_info:
                if self.comp.parent is None:
                    self.comp.endpoints.create(invalids, None)
                else:
                    with Instance(self.comp.parent) as parent:
                        self.comp.endpoints.create(invalids, parent)

    def test_create(self, count=1, exp_props=None):
        """Test the success and result of valid and invalid input to the create endpoint."""
        self.test_create_valid(count, exp_props)
        self.test_create_invalid(count)

    def test_update_valid(self, count=1, exp_props=None):
        """Test the success and result of valid input to the update endpoint.

        This test updates properties of an instance and validates
        that the updated and rerturned values are the same as the
        ones specified in valid payload data.
        You can limit the set of tested properties by specifying a list
        of expected property (exp_props), specified by their name.
        """
        print("# # # test update (valid)")
        for valids in self.comp.get_valids(Endpoints.UPDATE, count):
            exp_props_ = [prop.name for prop in self.comp.props.props if prop.spec_name in valids] if exp_props is None else exp_props
            print("valids:", valids)
            with Instance(self.comp) as instance:
                data = self.comp.endpoints.update(valids, instance)
                for name in exp_props_:
                    print("   ", name)
                    p = self.comp.props.get_prop(name)
                    assert self.comp.endpoints.res_has_prop(data, p.res_name)
                    assert valids[p.spec_name] == self.comp.endpoints.prop_from_res(data, p.res_name)

    def test_update_invalid(self, count=1):
        """Test the success and result of invalid input to the update endpoint.

        This test checks if the update endpoint indeed raises an exception
        if the provided data is invalid.
        """
        print("# # # test update (invalid)")
        for invalids in self.comp.get_invalids(Endpoints.UPDATE, count):
            print("invalids:", invalids)
            with Instance(self.comp) as instance:
                with pytest.raises(Exception) as e_info:
                    data = self.comp.endpoints.update(valids, instance)

    def test_update(self, count=1, exp_props=None):
        """Test the success and result of valid and invalid input to the update endpoint."""
        self.test_update_valid(count, exp_props)
        self.test_update_invalid(count)

    def test_get(self, count=1, times=2):
        """Test the success and result of the get endpoint.

        This test checks if the data of a component, obtained using the
        get endpoint, is the same as the one used during its creation.
        """
        print("# # # test get")
        for _ in range(times):
            with Instance(self.comp) as instance:
                print(instance.data)
                for valids in self.comp.get_valids(Endpoints.GET, count):
                    data = self.comp.endpoints.get(valids, instance)
                    for prop in self.comp.props.props:
                        k = prop.res_name
                        v = self.comp.endpoints.prop_from_res(instance.data, k)
                        assert self.comp.endpoints.res_has_prop(data, k)
                        assert v == self.comp.endpoints.prop_from_res(data, k)

    def test_list(self, count=1, entries=5, len_from_res=len):
        """Test the success of the list endpoint.

        This tests checks the count of entries obtained using the
        list endpoints.
        This is done by creating entries components and checking
        if the number of entries in list is actually increased
        accordingly.

        Parameters
        ----------
        count: int, default 1
            Count passed down to the valid data generator
        entries: int, default 5
            Number of entries to create during the test
        len_from_res: callable, default len
            Function to obtain the current component count from
            the list endpoint's result
        """
        print("# # # test list")
        for valids in self.comp.get_valids(Endpoints.LIST, count):
            initial = len_from_res(self.comp.endpoints.list(valids))
            print("initial:", initial)
            elements = []
            for i in range(entries):
                elements.append(Instance(self.comp, None))
                elements[-1].__enter__()
                current = len_from_res(self.comp.endpoints.list(valids))
                print("->", current)
                assert current == initial + i + 1
            for i in range(entries):
                elements[i].__exit__(None, None, None)
                current = len_from_res(self.comp.endpoints.list(valids))
                print("->", current)
                assert current == initial + entries - i - 1

        assert len(self.comp.endpoints.list(valids)) == initial

    def test_valid(self, key, count=1, val_res=None):
        """Test valid input to a non-standard endpoint.

        The custom endpoint is executed for all valid data
        in comp.get_valids(count).
        The result of this request is then validated using
        the custom function.

        Parameters
        ----------
        key: str
            Key to identify the endpoint (as specified
            during initialization of the Endpoints object)
        count: int, default 1
            Count passed down to the valid data generator
        val_res: callable, default None
            Function to validate the result from the
            custom endpoint
            It should return True is the result data is
            valid and False otherwise.
        """
        print("# # # test {} (valid)".format(key))
        for valids in self.comp.get_valids(key, count):
            print("valids:", valids)
            res = self.comp.endpoints.others[key](valids)
            print("        ->", res)
            if val_res is not None:
                assert(val_res(valids, res))

    def test_invalid(self, key, count=1):
        """Test invalid input to a non-standard endpoint.

        The custom endpoint is executed for all invalid data
        in comp.get_invalids(count).
        Then, the test checks if the endpoint's execution
        indeed raises an exception.

        Parameters
        ----------
        key: str
            Key to identify the endpoint (as specified
            during initialization of the Endpoints object)
        count: int, default 1
            Count passed down to the valid data generator
        """
        print("# # # test {} (invalid)".format(key))
        for invalids in self.comp.get_invalids(key, count):
            print("invalids:", invalids)
            with pytest.raises(Exception) as e_info:
                res = self.comp.endpoints.others[key](invalids)

    def test(self, key, count=1, val_res=None):
        """Test valid and invalid input to a non-standard endpoint."""
        self.test_valid(key, count, val_res)
        self.test_invalid(key, count)



import pytest
from .util import rand_int, rand_str, rand_uuid
from .property import Property
from .properties import Properties
from .endpoints import Endpoints
from .component import Component

def test_constr():
    p1 = Property("p1", rand_str, rand_int)
    p2 = Property("p2", rand_int, rand_str)
    def create(values, parent):
        return values
    def delete(values, instance):
        return instance.data
    endpoints = Endpoints(create=create, delete=delete)
    comp = Component("c1", Properties(p1, p2), endpoints)
    comp_test = ComponentTest(comp)



import json

def expand_dicts(dicts, key, values):
    """Expand and return the dicts with all values for the key.

    For each dict in dicts and each value in values, a new dict is generated.
    This new dict contains all entries from the original one as well
    as a mapping from key to value.
    If a value is None, a copy of the original dict is added.
    Overall, len(dicts)*len(values) dicts are generated and returned.

    Parameters
    ----------
    dicts: list
        List of dictionaries to expand
    key: str
        Key of the entry to be added to each dict
    values: list
        List of values to be added to each dict
    """
    expanded = []
    for old in dicts:
        for value in values:
            if value is not None:
                expanded.append({**old, key:value})
            else:
                expanded.append({**old})
    return expanded

def update_dicts(dicts, key, values):
    """Update and return the dicts with all values for the key.

    For each dict in dicts and each value in values, a new dict is generated.
    It contains all entries from the original one except for key.
    If a value is not None, key is mapped to value.
    Otherwise, key is not added to the dict as key.
    Overall, len(dicts)*len(values) dicts are generated and returned.

    Parameters
    ----------
    dicts: list
        List of dictionaries to update
    key: str
        Key of the entry to be updated in each dict
    values: list
        List of values to be added to each dict
    """
    updated = []
    for old in dicts:
        for value in values:
            old_ = {k:v for k,v in old.items() if k != key}
            if value is not None:
                old_.update({key: value})
            updated.append(old_)
    return updated

def dicts_without_key(dicts, key):
    """Return a list of new dicts without the key."""
    updated = []
    for old in dicts:
        updated.append({k:v for k,v in old.items() if k != key})
    return updated

def _dict_as_str(dict_):
    """Return as string representation of the dict."""
    return json.dumps(dict_, sort_keys=True)

def unique_dicts(dicts):
    """Return a list of all unique dicts."""
    return list({
        _dict_as_str(d):d
        for d in dicts
    }.values())

def exclude_dicts(dicts, exclude):
    """Return a list of all dicts that are not in exclude."""
    buts = [_dict_as_str(d) for d in exclude]
    return [
        d for d in dicts
        if _dict_as_str(d) not in buts
    ]

def non_empty_dicts(dicts):
    """Return a list all dicts with at least one entry."""
    return [x for x in dicts if len(x) > 0]

class Properties:
    """Wrapper for generating valid / invalid combinations of properties.

    Parameters
    ----------
    *props: list of Property
        List of the properties to wrap

    Examples
    --------
    prop1 = Property("prop1", rand_str, rand_int)
    prop2 = Property("prop2", rand_int, rand_str)
    props = Properties(prop1, prop2)
    """
    def __init__(self, *props):
        names = []
        spec_names = []
        res_names = []
        for prop in props:
            if prop.name in names:
                raise ValueError("Two properties with name '{}' specified.".format(prop.name))
            if prop.spec_name in spec_names:
                raise ValueError("Two properties with spec_name '{}' specified.".format(prop.spec_name))
            if prop.res_name in res_names:
                raise ValueError("Two properties with res_name '{}' specified.".format(prop.res_name))
            names.append(prop.name)
            spec_names.append(prop.spec_name)
            res_names.append(prop.res_name)
        self.props = props

    def get_prop(self, name=None, spec_name=None, res_name=None):
        if name is None and spec_name is None and res_name is None:
            raise ValueError("Must specify a name to get prop for.")
        print(name, spec_name, res_name)
        for prop in self.props:
            print(prop.name, prop.spec_name, prop.res_name)
            if name is not None and prop.name == name:
                return prop
            elif spec_name is not None and prop.spec_name == spec_name:
                return prop
            elif res_name is not None and prop.res_name == res_name:
                return prop
        if name is not None:
            raise KeyError("No prop exists with name {}.".format(name))
        if spec_name is not None:
            raise KeyError("No prop exists with spec_name {}.".format(spec_name))
        if res_name is not None:
            raise KeyError("No prop exists with res_name {}.".format(res_name))

    def get_valids(self, method, count=1):
        """Return a list of valid value dicts from all properties.

        For each property, a list of valid values is generated.
        Then, all valid combinations of these values are generated
        and stored in dicts.
        As a key, the name of a property is used.

        Parameters
        ----------
        endpoint: str
            Key of the endpoint to get values for
        count: int
            Number of values to generate per property
        """
        all_valids = [{}]
        for prop in self.props:
            valids = prop.get_valids(method, count)
            if len(valids) > 0:
                all_valids = expand_dicts(all_valids, prop.spec_name, valids)
        return all_valids

    def get_invalids(self, method, count=1):
        """Return a list of invalid value dicts from all properties.

        Parameters
        ----------
        endpoint: str
            Key of the endpoint to get values for
        count: int
            Number of values to generate per property
        """
        all_valids = self.get_valids(method, count)
        all_invalids = []
        # create invalids from valids
        for prop in self.props:
            invalids = prop.get_invalids(method, count)
            if len(invalids) > 0:
                updated = [{k:v for k,v in x.items()} for x in all_valids]
                updated = update_dicts(updated, prop.spec_name, invalids)
                all_invalids += updated
        # remove duplicated
        all_invalids = unique_dicts(all_invalids)
        # remove empty dict
        all_invalids = non_empty_dicts(all_invalids)
        # add empty dict if this would be invalid
        for prop in self.props:
            if prop.is_required(method):
                all_invalids.append({})
                break
        # remove still valid dicts
        all_invalids = exclude_dicts(all_invalids, all_valids)
        return all_invalids


import pytest
from .property import Property
from .endpoints import Endpoints
from .util import rand_str

def test_names():
    p1 = Property(rand_str(), [], [], spec_name=rand_str(), res_name=rand_str())
    p2 = Property(rand_str(), [], [], spec_name=rand_str(), res_name=rand_str())
    p1_1 = Property(p1.name, [], [], spec_name=rand_str(), res_name=rand_str())
    p1_2 = Property(rand_str(), [], [], spec_name=p1.spec_name, res_name=rand_str())
    p1_3 = Property(rand_str(), [], [], spec_name=rand_str(), res_name=p1.res_name)
    Properties(p1, p2)
    Properties(p2, p1_1)
    Properties(p2, p1_2)
    Properties(p2, p1_3)
    with pytest.raises(Exception):
        Properties(p1, p1_1)
    with pytest.raises(Exception):
        Properties(p1, p1_2)
    with pytest.raises(Exception):
        Properties(p1, p1_3)

def test_prop_by_name():
    p1 = Property(rand_str(), [], [], spec_name=rand_str(), res_name=rand_str())
    p2 = Property(rand_str(), [], [], spec_name=rand_str(), res_name=rand_str())
    props = Properties(p1, p2)
    assert props.get_prop(p1.name).name == p1.name
    assert props.get_prop(name=p1.name).name == p1.name
    assert props.get_prop(spec_name=p1.spec_name).name == p1.name
    assert props.get_prop(res_name=p1.res_name).name == p1.name
    with pytest.raises(Exception):
        props.get_prop(rand_str())
    with pytest.raises(Exception):
        props.get_prop(name=rand_str())
    with pytest.raises(Exception):
        props.get_prop(spec_name=rand_str())
    with pytest.raises(Exception):
        props.get_prop(res_name=rand_str())

def test_valids():
    c, u = Endpoints.CREATE, Endpoints.UPDATE
    p1 = Property("p1", [1, 2, 3, 4, 5], [], [c])
    p2 = Property("p2", [11, 12, 13], [])
    props = Properties(p1, p2)
    assert len(props.get_valids(c, count=1)) == len(p1.get_valids(c)) * len(p2.get_valids(c))
    assert len(props.get_valids(u, count=1)) == len(p1.get_valids(u)) * len(p2.get_valids(u))



import random
from .endpoints import Endpoints

class Instance:
    """Wrapper for the creation and deletion of a component.

    An instance object provides __enter__ and __exit__ methods.
    They use the component's andpoints specification to
    create a resource and delete it.

    During enter, a component is created using the endpoint
    comp.endpoints.create.
    As payload data, it uses a random element from the
    generated list of valids for comp.endpoints.CREATE.

    During exit, the component is deleted using the endpoint
    comp.endpooints.delete.
    As payload data, it uses a random element from the
    generated list of valids for comp.endpoints.DELETE.

    The result of the create endpoint is stored in
    self.data for accessing it later.

    Parameters
    ----------
    comp: Component object
        Component to create an instance from
    parent: Instance object, default None
        Instance of a parent (as specified by the component's parent)
    **values: arbitrary
        Arbitrary values to overwrite values (from valids) for resource creation

    Examples
    --------
    comp = ...
    with Instance(comp) as instance:
        print(instance.data)
    """
    def __init__(self, comp, parent=None, **values):
        self.comp = comp
        self.parent = parent
        self.values = values

    def __enter__(self):
        if self.comp.parent is not None and self.parent is None:
            self.parent = Instance(self.comp.parent)
            self.parent.__enter__()
            self.parent_created = True
        else:
            self.parent_created = False

        valids = random.choice(self.comp.get_valids(Endpoints.CREATE))
        values_ = {**valids, **self.values}
        self.data = self.comp.endpoints.create(values_, self.parent)
        return self

    def __exit__(self, type, value, traceback):
        valids = random.choice(self.comp.get_valids(Endpoints.DELETE))
        self.comp.endpoints.delete(valids, self)
        if self.parent_created:
            self.parent.__exit__(None, None, None)


import pytest
from .util import rand_int, rand_str, rand_uuid
from .property import Property
from .properties import Properties
from .endpoints import Endpoints
from .component import Component

def test_constr():
    p1 = Property("p1", rand_str, rand_int)
    p2 = Property("p2", rand_int, rand_str)
    def create(values, parent):
        return values
    def delete(values, instance):
        return instance.data
    endpoints = Endpoints(create=create, delete=delete)
    comp = Component("c1", Properties(p1, p2), endpoints)
    with Instance(comp) as instance:
        pass



import random

from apitest import Property, Properties, Endpoints, Component, Instance, ComponentTest
from apitest.util import rand_str, rand_int, value_from_dict, dict_has_value


class CategoryApi:
    def __init__(self):
        self.categories = {}

    def create(self, values, parent):
        if "name" not in values:
            raise ValueError("must specify name")
        if values["name"] in [None, ""]:
            raise ValueError("name cannot be empty")
        id = random.randint(0, 1000)
        data = {**values, "id": id}
        self.categories[id] = data
        return data

    def update(self, values, instance):
        id = instance.data["id"]
        data = {**self.categories[id], **values}
        self.categories[instance.data["id"]] = data
        return data

    def delete(self, values, instance):
        id = instance.data["id"]
        data = self.categories[id]
        del self.categories[id]
        return data

    def list(self, values):
        return [v for v in self.categories.values()]

class EntryApi:
    def __init__(self):
        self.entries = {}

    def create(self, values, parent):
        if "name" not in values:
            raise ValueError("must specify name")
        if values["name"] in [None, ""]:
            raise ValueError("name cannot be empty")
        if "color" not in values:
            raise ValueError("must specify color")
        if values["color"] not in ["red", "green", "blue"]:
            raise ValueError("color is invalid: {}".format(values["color"]))
        id = random.randint(1, 1000)
        data = {**values, "category_id": parent.data["id"], "id": id}
        self.entries[id] = data
        return data

    def update(self, values, instance):
        id = instance.data["id"]
        data = {**self.entries[id], **values}
        self.entries[instance.data["id"]] = data
        return data

    def delete(self, values, instance):
        id = instance.data["id"]
        data = self.entries[id]
        del self.entries[id]
        return data

    def list(self, values):
        return [v for v in self.entries.values()]

class ComputeApi:
    def compute(self, data):
        return data["counter"] + 1



capi = CategoryApi()
category_api = Endpoints(
    create=capi.create,
    update=capi.update,
    delete=capi.delete,
    list=capi.list,
    prop_from_res=value_from_dict,
    res_has_prop=dict_has_value
)

eapi = EntryApi()
entry_api = Endpoints(
    create=eapi.create,
    update=eapi.update,
    delete=eapi.delete,
    list=eapi.list,
    prop_from_res=value_from_dict,
    res_has_prop=dict_has_value
)

compapi = ComputeApi()
compute_api = Endpoints(
    compute=compapi.compute
)

name = Property("name", rand_str, [""], [Endpoints.CREATE])
description = Property("description", rand_str, [])
color = Property("color", ["red", "green", "blue"], rand_str, [Endpoints.CREATE])
counter = Property("counter", rand_int, rand_str, ["compute"])

category = Component("Category", Properties(name, description), category_api)
entry = Component("Entry", Properties(name, description, color), entry_api, category)
computation = Component("Computation", Properties(counter), compute_api)

def test_create_category():
    ComponentTest(category).test_create()
def test_update_category():
    ComponentTest(category).test_update()
def test_list_category():
    ComponentTest(category).test_list(entries=5)

def test_create_entry():
    ComponentTest(entry).test_create()
def test_update_entry():
    ComponentTest(entry).test_update()
def test_list_entry():
    ComponentTest(entry).test_list(entries=5)

def test_compute():
    ComponentTest(computation).test("compute", 2, val_res=lambda values,x: type(x) == int)



from .util import attr_from_obj, obj_has_attr

class Endpoints:
    """Specification of endpoints of an API.

    The following endpoints with specific meaning can be specified:

    1. create, signature: create(values, parent)
    2. update, signature: update(values, instance)
    3. delete, signature: delete(values, instance)
    4. get, signature: get(values, instance)
    5. list, signature: list(values)

    These endpoints are identified with the following keys:

    1. create: Endpoints.CREATE
    2. update: Endpoints.UPDATE
    2. delete: Endpoints.DELETE
    2. get: Endpoints.GET
    2. list: Endpoints.LIST

    Other endpoints can be specified with arbitrary names.
    They must have the following signature:

    6. other, signature: other(values)

    Each of these endpoints is identified with its name as key,
    e.g., "other".

    Parameters
    ----------
    create: callable, default None
        Function for creating a new resource
    update: callable, default None
        Function for updating an existing resource
    delete: callable, default None
        Function for deleting an existing resource
    get: callable, default None
        Function for getting an existing resource
    list: callable, default None
        Function to list all existing resources
    prop_from_res: callable, default None
        Function to get a property from an API call result
    res_has_prop: callable, default None
        Function to determine if an API call result has a property
    **others: callable
        Other functions

    Examples
    --------
    component_api = Endpoints(
        create=lambda values,parent: {**values, "parent_id": parent.data.id},
        update=lambda values,instance: {**instance.data, **values},
        delete=lambda values,instance: {**instance.data},
        get=lambda values,instance: {**instance.data},
        list=lambda values: [],
        prop_from_res=value_from_dict,
        res_has_prop=dict_has_value
    )
    compute_api = Endppoints(
        square=lambda values: values["number"]**2
    )
    """

    CREATE = "CREATE"
    UPDATE = "UPDATE"
    DELETE = "DELETE"
    GET = "GET"
    LIST = "LIST"

    def __init__(self,
                 create=None,
                 update=None,
                 delete=None,
                 get=None,
                 list=None,
                 prop_from_res=attr_from_obj,
                 res_has_prop=obj_has_attr,
                 **others):
        self.create = create
        self.get = get
        self.update = update
        self.delete = delete
        self.list = list
        self.others = others
        self.prop_from_res = prop_from_res
        self.res_has_prop = res_has_prop


import pytest

def test_constr():
    Endpoints()
    Endpoints(create=lambda values,parent: 1)
    Endpoints(update=lambda values,instance: 2, delete=lambda values,instance: 3)
    Endpoints(xyz=lambda values,parent: 4)
